using System;
using System.Text;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Utilities.Test;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Utilities.IO;

using bcfipsin100.util;

public class AesExamples
{
 
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("==========================");
        Console.WriteLine("Start Of AesExamples.Main");
        Console.WriteLine("==========================");

        byte[] cipherTextData;
        byte[][] cipherMacData;
        byte[] plainTextData;
        byte[][] plainMacData;
        byte[] macTextData; 

        /// <summary>
        /// Generating a secure key which can be used for symmetric encryption
        /// </summary>
        SymmetricSecretKey secKey1 = generateKey();
        Console.WriteLine(string.Format("secKey1:[{0}]",string.Join(",",secKey1.GetKeyBytes())));
        

        /// <summary>
        /// Generating a symmetric key from a your own input as a key
        /// </summary>
        SymmetricSecretKey unsecKey1 = defineKey(new byte[] {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15});
        Console.WriteLine(string.Format("unsecKey1 128bits:[{0}]",string.Join(",",unsecKey1.GetKeyBytes())));
        SymmetricSecretKey unsecKey2 = defineKey(new byte[] {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23});
        Console.WriteLine(string.Format("unsecKey2 192bits:[{0}]",string.Join(",",unsecKey2.GetKeyBytes())));
        SymmetricSecretKey unsecKey3 = defineKey(new byte[] {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31});
        Console.WriteLine(string.Format("unsecKey3 256bits:[{0}]",string.Join(",",unsecKey3.GetKeyBytes())));
        Console.WriteLine();
        
        /// <summary>
        /// Showing how to encrypt and decrypt using AES ECB Mode
        /// </summary>
        cipherTextData = ecbEncrypt(ExValues.sampleAesKey,ExValues.sampleInput.Concat(
            Encoding.Default.GetBytes(" AES ECB")).ToArray());
        plainTextData = ecbDecrypt(ExValues.sampleAesKey,cipherTextData);
        Console.WriteLine(string.Format("ECB encryption/decryption    plainTextData=[{0}]       decryptedText=[{1}]",
           Encoding.Default.GetString(ExValues.sampleInput) + " AES ECB",
           Encoding.Default.GetString(plainTextData)));
        

        /// <summary>
        /// Showing how to encrypt and decrypt using AES CBC Mode
        /// </summary>
        cipherTextData = cbcEncrypt(ExValues.sampleAesKey,ExValues.sampleInput.Concat(
            Encoding.Default.GetBytes(" AES CBC")).ToArray());
        plainTextData = cbcDecrypt(ExValues.sampleAesKey,cipherTextData);
        Console.WriteLine(string.Format("CBC encryption/decryption    plainTextData=[{0}]       decryptedText=[{1}]",
           Encoding.Default.GetString(ExValues.sampleInput) + " AES CBC",
           Encoding.Default.GetString(plainTextData)));

        /// <summary>
        /// Showing how to encrypt and decrypt using AES CBC Mode with auto generating a 
        /// secure IV
        /// </summary>        
        byte[][] result;
        result = cbcEncryptSecureIV(ExValues.sampleAesKey,ExValues.sampleInput.Concat(
            Encoding.Default.GetBytes(" CBC SECURE-IV")).ToArray());        
        plainTextData = cbcDecryptSecureIV(ExValues.sampleAesKey,result[0],result[1]);
        Console.WriteLine(string.Format("CBC with Secure Random IV    plainTextData=[{0}] decryptedText=[{1}]",
           Encoding.Default.GetString(ExValues.sampleInput) + " CBC SECURE-IV",
           Encoding.Default.GetString(plainTextData)));

        /// <summary>
        /// Showing how to encrypt and decrypt using AES CFB Mode
        /// </summary>
        cipherTextData = cfbEncrypt(ExValues.sampleAesKey,ExValues.sampleInput.Concat(
            Encoding.Default.GetBytes(" AES CFB")).ToArray());
        plainTextData = cfbDecrypt(ExValues.sampleAesKey,cipherTextData);
        Console.WriteLine(string.Format("CFB encryption/decryption    plainTextData=[{0}]       decryptedText=[{1}]",
           Encoding.Default.GetString(ExValues.sampleInput) + " AES CFB",
           Encoding.Default.GetString(plainTextData)));

        /// <summary>
        /// Showing how to encrypt and decrypt using AES CTR Mode
        /// </summary>
        cipherTextData = ctrEncrypt(ExValues.sampleAesKey,ExValues.sampleInput.Concat(
            Encoding.Default.GetBytes(" AES CTR")).ToArray());
        plainTextData = ctrDecrypt(ExValues.sampleAesKey,cipherTextData);
        Console.WriteLine(string.Format("CTR encryption/decryption    plainTextData=[{0}]       decryptedText=[{1}]",
           Encoding.Default.GetString(ExValues.sampleInput) + " AES CTR",
           Encoding.Default.GetString(plainTextData)));
           

        /// <summary>
        /// Showing how to encrypt and decrypt using AES CBC Mode with Cipher Text Stealing (CTS)
        /// </summary>
        cipherTextData = ctsWithCS3Encrypt(ExValues.sampleAesKey,ExValues.sampleInput.Concat(
            Encoding.Default.GetBytes(" AES CTS")).ToArray());
        plainTextData = ctsWithCS3Decrypt(ExValues.sampleAesKey,cipherTextData);
        Console.WriteLine(string.Format("CTS encryption/decryption    plainTextData=[{0}]       decryptedText=[{1}]",
           Encoding.Default.GetString(ExValues.sampleInput) + " AES CTS",
           Encoding.Default.GetString(plainTextData)));


        /// <summary>
        /// Showing how to encrypt and decrypt using AES GCM Mode NO AAD
        /// </summary>
        cipherTextData = gcmEncrypt(ExValues.sampleAesKey,ExValues.sampleInput.Concat(
            Encoding.Default.GetBytes(" GCM NO AAD")).ToArray());
        plainTextData = gcmDecrypt(ExValues.sampleAesKey,cipherTextData);
        macTextData = ExValues.SubArray(cipherTextData,cipherTextData.Length-16,16);   
        Console.WriteLine(string.Format("GCM encryption/decryption    plainTextData=[{0}]    decryptedText=[{1}]   MAC=[{2}]",
           Encoding.Default.GetString(ExValues.sampleInput) + " GCM NO AAD",
           Encoding.Default.GetString(plainTextData),
           string.Join(",",macTextData)));

        /// <summary>
        /// Showing how to encrypt and decrypt using AES GCM Mode with AAD
        /// </summary>
        cipherMacData = gcmEncryptAAD(ExValues.sampleAesKey,ExValues.sampleInput.Concat(
            Encoding.Default.GetBytes(" GCM WITH AAD")).ToArray());
        plainMacData = gcmDecryptAAD(ExValues.sampleAesKey,cipherMacData[0]);
        Console.WriteLine(string.Format("GCM encryption/decryption    plainTextData=[{0}]  decryptedText=[{1}] MAC=[{2}]",
           Encoding.Default.GetString(ExValues.sampleInput) + " GCM WITH ADD",
           Encoding.Default.GetString(plainMacData[0]),
           string.Join(",",cipherMacData[1])));
        //Console.WriteLine(string.Format("GCM encryption/decryption    cipher$$$$Data=[{0}]",string.Join(",",cipherMacData[0])));

        /// <summary>
        /// Showing how to encrypt and decrypt using AES CCM
        /// </summary>
        cipherTextData = ccmEncrypt(ExValues.sampleAesKey,ExValues.sampleInput.Concat(
            Encoding.Default.GetBytes(" CCM NO AAD")).ToArray());
        plainTextData = ccmDecrypt(ExValues.sampleAesKey,cipherTextData);
        macTextData = ExValues.SubArray(cipherTextData,cipherTextData.Length-16,16);   
        Console.WriteLine(string.Format("CCM encryption/decryption    plainText=[{0}]        decryptedText=[{1}]   MAC=[{2}]",
           Encoding.Default.GetString(ExValues.sampleInput) + " CCM NO AAD",
           Encoding.Default.GetString(plainTextData),
           string.Join(",",macTextData)));
           

        /// <summary>
        /// Showing how to encrypt and decrypt using AES CCM WITH AAD
        /// </summary>
        cipherTextData = ccmEncryptAAD(ExValues.sampleAesKey,ExValues.sampleInput.Concat(
                                       Encoding.Default.GetBytes(" CCM WITH AAD")).ToArray());
        plainTextData = ccmDecryptAAD(ExValues.sampleAesKey,cipherTextData);
        Console.WriteLine(string.Format("CCM encryption/decryption    plainText=[{0}]      decryptedText=[{1}] AAD=[{2}]",
           Encoding.Default.GetString(ExValues.sampleInput.Concat(
                                       Encoding.Default.GetBytes(" CCM WITH AAD")).ToArray()),
           Encoding.Default.GetString(plainTextData),
           Encoding.Default.GetString(ExValues.additionalAuthenticatedDataA)));


        /// <summary>
        /// Showing how to wrap input bytes using AES without padding
        /// </summary>
        cipherTextData = wrapInputWithoutPadding(ExValues.sampleAesKey, ExValues.sampleUnpaddedKey);
        Console.WriteLine(string.Format("AES Key Unpadded Wrapping    AESkey=[{0}]        originalInput=[{1}]    wrappedkey=[{2}]",
           string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(0,20) + "...," + string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(35,2),
           Encoding.Default.GetString(ExValues.sampleUnpaddedKey),
           string.Join(",",cipherTextData).Substring(0,43) + "..."));
        plainTextData = unwrapInputWithoutPadding(ExValues.sampleAesKey, cipherTextData);
        Console.WriteLine(string.Format("AES Key Unpadded Unwrapping  AESkey=[{0}]        originalInput=[{1}]    unwrappedkey=[{2}]",
           string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(0,20) + "...," + string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(35,2),
           Encoding.Default.GetString(ExValues.sampleUnpaddedKey),
           Encoding.Default.GetString(plainTextData)));


        /// <summary>
        /// Showing how to wrap input bytes using AES with padding
        /// </summary>
        cipherTextData = wrapInputWithPadding(ExValues.sampleAesKey, ExValues.samplePaddedKey);
        Console.WriteLine(string.Format("AES Key Padded Wrapping      AESkey=[{0}]        originalInput=[{1}] wrappedkey=[{2}]",
           string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(0,20) + "...," + string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(35,2),
           Encoding.Default.GetString(ExValues.samplePaddedKey),
           string.Join(",",cipherTextData).Substring(0,44) + "..."));
        plainTextData = unwrapInputWithPadding(ExValues.sampleAesKey, cipherTextData);
        Console.WriteLine(string.Format("AES Key Padded Unwrapping    AESkey=[{0}]        originalInput=[{1}] unwrappedkey=[{2}]",
           string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(0,20) + "...," + string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(35,2),
           Encoding.Default.GetString(ExValues.samplePaddedKey),
           Encoding.Default.GetString(plainTextData)));


        /// <summary>
        /// Showing how to wrap AES key using AES without padding
        /// </summary>
        cipherTextData = wrapKeyWithoutPadding(ExValues.sampleAesKey, secKey1);
        Console.WriteLine(string.Format("AES Key Unpadded Wrapping    AESkey=[{0}]        originalkey=[{1}]    wrappedkey=[{2}]",
           string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(0,20) + "...," + string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(35,2),
           string.Join(",",secKey1.GetKeyBytes()).Substring(0,20) + "...," + string.Join(",",secKey1.GetKeyBytes()).Substring(35,2),
           string.Join(",",cipherTextData).Substring(0,43) + "..."));
        SymmetricSecretKey secKey2 = unwrapKeyWithoutPadding(ExValues.sampleAesKey, cipherTextData);
        Console.WriteLine(string.Format("AES Key Unpadded Unwrapping  AESkey=[{0}]        originalkey=[{1}]    unwrappedkey=[{2}]",
           string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(0,20) + "...," + string.Join(",",ExValues.sampleAesKey.GetKeyBytes()).Substring(35,2),
           string.Join(",",secKey1.GetKeyBytes()).Substring(0,20) + "...," + string.Join(",",secKey1.GetKeyBytes()).Substring(35,2),
           string.Join(",",secKey2.GetKeyBytes()).Substring(0,20) + "...," + string.Join(",",secKey2.GetKeyBytes()).Substring(35,2)));


        Console.WriteLine("==========================");
        Console.WriteLine("End   Of AesExamples.Main");
        Console.WriteLine("==========================");
        Console.WriteLine();
        Console.WriteLine();        
    }
    
    
    /// <summary>
    /// Example 1 – Creating a FIPS Approved SecureRandom
    /// </summary>
    public static FipsSecureRandom buildDrbg()
    {   
        BasicEntropySourceProvider entSource = new BasicEntropySourceProvider(new SecureRandom(), true);
        IDrbgBuilder<FipsSecureRandom> theDrbgBuilder = CryptoServicesRegistrar.CreateService(
                                                    FipsDrbg.Sha512HMac).FromEntropySource(entSource);                
        theDrbgBuilder.SetSecurityStrength(256);
        theDrbgBuilder.SetEntropyBitsRequired(256);
            
        return theDrbgBuilder.Build(ExValues.sampleNonce,false);
    } 

    
    /// <summary>
    /// Example 4 – Generating an AES Key
    /// </summary>
    public static SymmetricSecretKey generateKey()
    {
        FipsAes.KeyGenerator myKeyGen = CryptoServicesRegistrar.CreateGenerator(
                                        FipsAes.KeyGen256, new SecureRandom());
        return myKeyGen.GenerateKey();
    }


    /// <summary>
    /// Example 5 – Key Construction with a given SecretKey
    /// </summary>
    public static SymmetricSecretKey defineKey(byte[] keyBytes)
    {
        FipsAes.Key myAESKey = new FipsAes.Key(keyBytes);
        return myAESKey; 
    }
    
    
    /// <summary>
    /// Example 6 – ECB Mode Encryption
    /// </summary>    
    public static byte[] ecbEncrypt(FipsAes.Key key, byte[] plainTextData)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IBlockCipherBuilder<IParameters<Algorithm>> encryptBuilder =
                    provider.CreateBlockEncryptorBuilder(FipsAes.Ecb);

        MemoryOutputStream streamOutput = new MemoryOutputStream();
        ICipher pkcs7PaddedEncryptor = encryptBuilder.BuildPaddedCipher(streamOutput,new Pkcs7Padding());
                    
        using (Stream encrypterStream = pkcs7PaddedEncryptor.Stream)
        { encrypterStream.Write(plainTextData,0,plainTextData.Length); }    
                
        byte[] cipherTextData = streamOutput.ToArray();
        return cipherTextData;
    }    
    

    /// <summary>
    /// Example 6 – ECB Mode Deccryption
    /// </summary>
    public static byte[] ecbDecrypt(FipsAes.Key key, byte[] cipherTextData)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IBlockCipherBuilder<IParameters<Algorithm>> decryptBuilder =
                    provider.CreateBlockDecryptorBuilder(FipsAes.Ecb);

        ICipher pkcs7PaddedDecryptor = decryptBuilder.BuildPaddedCipher(
               new MemoryInputStream(cipherTextData),
               new Pkcs7Padding());
                    
        byte[] plainTextData;
        using (Stream decrypterStream = pkcs7PaddedDecryptor.Stream)
        { plainTextData = Streams.ReadAll(decrypterStream); }    
                
        return plainTextData;
    } 


    /// <summary>
    /// Example 7 – CBC Mode Encryption
    /// </summary>    
    public static byte[] cbcEncrypt(FipsAes.Key key, byte[] plainTextData)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
               
        IBlockCipherBuilder<IParameters<Algorithm>> encryptBuilder =
                    provider.CreateBlockEncryptorBuilder(FipsAes.Cbc.WithIV(ExValues.sampleIVnonce));
               
        MemoryOutputStream streamOutput = new MemoryOutputStream();
        ICipher pkcs7PaddedEncryptor = encryptBuilder.BuildPaddedCipher(streamOutput,new Pkcs7Padding());
                    
        using (Stream encrypterStream = pkcs7PaddedEncryptor.Stream)
        { encrypterStream.Write(plainTextData,0,plainTextData.Length); }    
                
        byte[] cipherTextData = streamOutput.ToArray();
        return cipherTextData;
    }

    /// <summary>
    /// Example 7 – CBC Mode Deccryption
    /// </summary>    
    public static byte[] cbcDecrypt(FipsAes.Key key, byte[] cipherTextData)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IBlockCipherBuilder<IParameters<Algorithm>> decryptBuilder =
                    provider.CreateBlockDecryptorBuilder(FipsAes.Cbc.WithIV(ExValues.sampleIVnonce));

        ICipher pkcs7PaddedDecryptor = decryptBuilder.BuildPaddedCipher(
               new MemoryInputStream(cipherTextData),
               new Pkcs7Padding());
                    
        byte[] plainTextData;
        using (Stream decrypterStream = pkcs7PaddedDecryptor.Stream)
        { plainTextData = Streams.ReadAll(decrypterStream); }    
                
        return plainTextData;
    } 
  
  
  
    /// <summary>
    /// Example 7a – CBC Mode Encryption with auto generated secure IV
    /// </summary>    
    public static byte[][] cbcEncryptSecureIV(FipsAes.Key key, byte[] plainTextData)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
               
        IBlockCipherBuilder<IParameters<Algorithm>> encryptBuilder =
                    provider.CreateBlockEncryptorBuilder(FipsAes.Cbc.WithIV(buildDrbg()));
       
        FipsAes.ParametersWithIV  cbcIV = (FipsAes.ParametersWithIV)encryptBuilder.AlgorithmDetails;
        
        MemoryOutputStream streamOutput = new MemoryOutputStream();
        ICipher pkcs7PaddedEncryptor = encryptBuilder.BuildPaddedCipher(streamOutput,new Pkcs7Padding());
                    
        using (Stream encrypterStream = pkcs7PaddedEncryptor.Stream)
        { encrypterStream.Write(plainTextData,0,plainTextData.Length); }    
        
        byte[] cipherTextData = streamOutput.ToArray();
        byte[] cipherIV = cbcIV.GetIV(); 
        byte[][] cipherTextDataWithIV = {cipherTextData, cipherIV}; 
        return cipherTextDataWithIV;
    }

    /// <summary>
    /// Example 7a – CBC Mode Deccryption where we pass the previously auto
    /// generated IV
    /// </summary>    
    public static byte[] cbcDecryptSecureIV(FipsAes.Key key, byte[] cipherTextData, byte[] cbcIV)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IBlockCipherBuilder<IParameters<Algorithm>> decryptBuilder =
                    provider.CreateBlockDecryptorBuilder(FipsAes.Cbc.WithIV(cbcIV));

        ICipher pkcs7PaddedDecryptor = decryptBuilder.BuildPaddedCipher(
               new MemoryInputStream(cipherTextData),
               new Pkcs7Padding());
                    
        byte[] plainTextData;
        using (Stream decrypterStream = pkcs7PaddedDecryptor.Stream)
        { plainTextData = Streams.ReadAll(decrypterStream); }    
                
        return plainTextData;
    } 
    
    
    /// <summary>
    /// Example 8 – CFB Mode Encryption
    /// </summary>    
    public static byte[] cfbEncrypt(FipsAes.Key key, byte[] plainTextData)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        ICipherBuilder<IParameters<Algorithm>> encryptBuilder =
                    provider.CreateEncryptorBuilder(FipsAes.Cfb128.WithIV(ExValues.sampleIVnonce));

        MemoryOutputStream streamOutput = new MemoryOutputStream();
        ICipher noPaddedEncryptor = encryptBuilder.BuildCipher(streamOutput);
                    
        using (Stream encrypterStream = noPaddedEncryptor.Stream)
        { encrypterStream.Write(plainTextData,0,plainTextData.Length); }    
                
        byte[] cipherTextData = streamOutput.ToArray();
        return cipherTextData;
    }    
    
    /// <summary>
    /// Example 8 – CFB Mode Decryption
    /// </summary>    
    public static byte[] cfbDecrypt(FipsAes.Key key, byte[] cipherTextData)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        ICipherBuilder<IParameters<Algorithm>> decryptBuilder =
                    provider.CreateDecryptorBuilder(FipsAes.Cfb128.WithIV(ExValues.sampleIVnonce));

        ICipher noPaddedDecryptor = decryptBuilder.BuildCipher(
                    new MemoryInputStream(cipherTextData));
                    
        byte[] plainTextData;
        using (Stream decrypterStream = noPaddedDecryptor.Stream)
        { plainTextData = Streams.ReadAll(decrypterStream); }    
                
        return plainTextData;
    }    
    
    
    /// <summary>
    /// Example 9 – CTR Mode Encryption
    /// </summary>    
    public static byte[] ctrEncrypt(FipsAes.Key key, byte[] plainTextData)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        ICipherBuilder<IParameters<Algorithm>> encryptBuilder =
                    provider.CreateEncryptorBuilder(FipsAes.Ctr.WithIV(ExValues.sampleIVcounter));

        MemoryOutputStream streamOutput = new MemoryOutputStream();
        ICipher noPaddedEncryptor = encryptBuilder.BuildCipher(streamOutput);
                    
        using (Stream encrypterStream = noPaddedEncryptor.Stream)
        { encrypterStream.Write(plainTextData,0,plainTextData.Length); }    
                
        byte[] cipherTextData = streamOutput.ToArray();
        return cipherTextData;
    }

    /// <summary>
    /// Example 9 – CTR Mode Decryption
    /// </summary>    
    public static byte[] ctrDecrypt(FipsAes.Key key, byte[] cipherTextData)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        ICipherBuilder<IParameters<Algorithm>> decryptBuilder =
                    provider.CreateDecryptorBuilder(FipsAes.Ctr.WithIV(ExValues.sampleIVcounter));

        ICipher noPaddedDecryptor = decryptBuilder.BuildCipher(
                    new MemoryInputStream(cipherTextData));
                    
        byte[] plainTextData;
        using (Stream decrypterStream = noPaddedDecryptor.Stream)
        { plainTextData = Streams.ReadAll(decrypterStream); }    
                
        return plainTextData;
    }    
    
    
    /// <summary>
    /// Example 10 – CTS Mode Encryption With CS3
    /// </summary>    
    public static byte[] ctsWithCS3Encrypt(FipsAes.Key key, byte[] plainTextData)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IBlockCipherBuilder<IParameters<Algorithm>> encryptBuilder =
                    provider.CreateBlockEncryptorBuilder(
                        FipsAes.CbcWithCS3.WithIV(ExValues.sampleIVnonce));

        MemoryOutputStream streamOutput = new MemoryOutputStream();
        ICipher pkcs7PaddedEncryptor = encryptBuilder.BuildPaddedCipher(streamOutput,new Pkcs7Padding());
                    
        using (Stream encrypterStream = pkcs7PaddedEncryptor.Stream)
        { encrypterStream.Write(plainTextData,0,plainTextData.Length); }    
                
        byte[] cipherTextData = streamOutput.ToArray();
        return cipherTextData;
    }

    /// <summary>
    /// Example 10 – CTS Mode Decryption With CS3
    /// </summary>    
    public static byte[] ctsWithCS3Decrypt(FipsAes.Key key, byte[] cipherTextData)
    {
        IBlockCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IBlockCipherBuilder<IParameters<Algorithm>> decryptBuilder =
                    provider.CreateBlockDecryptorBuilder(
                        FipsAes.CbcWithCS3.WithIV(ExValues.sampleIVnonce));

        ICipher pkcs7PaddedDecryptor = decryptBuilder.BuildPaddedCipher(
               new MemoryInputStream(cipherTextData),
               new Pkcs7Padding());
                    
        byte[] plainTextData;
        using (Stream decrypterStream = pkcs7PaddedDecryptor.Stream)
        { plainTextData = Streams.ReadAll(decrypterStream); }    
                
        return plainTextData;
    }
    
    
    /// <summary>
    /// Example 11 – GCM Mode Encryption with AAD
    /// </summary>    
    public static byte[][] gcmEncryptAAD(FipsAes.Key key, byte[] plainTextData)
    {        
        IAeadCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IAeadCipherBuilder<IParameters<Algorithm>> encryptBuilder =
                    provider.CreateAeadEncryptorBuilder(FipsAes.Gcm.WithIV(ExValues.sampleIVcounter));

        MemoryOutputStream streamOutput = new MemoryOutputStream();
        
        IAeadCipher noPaddedEncryptor = encryptBuilder.BuildAeadCipher(AeadUsage.AAD_FIRST, streamOutput);
        noPaddedEncryptor.AadStream.Write(ExValues.additionalAuthenticatedDataA,0,
                                          ExValues.additionalAuthenticatedDataA.Length);
                                          
        using (Stream encrypterStream = noPaddedEncryptor.Stream)
        { encrypterStream.Write(plainTextData,0,plainTextData.Length); }    
          
          
        byte[][] cipherMacData = { streamOutput.ToArray(), noPaddedEncryptor.GetMac().Collect() };
        
        return cipherMacData;
    }
    
    /// <summary>
    /// Example 11 – GCM Mode Decryption with AAD
    /// </summary>    
    public static byte[][] gcmDecryptAAD(FipsAes.Key key, byte[] cipherTextData)
    {        
        IAeadCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IAeadCipherBuilder<IParameters<Algorithm>> decryptBuilder =
                    provider.CreateAeadDecryptorBuilder(FipsAes.Gcm.WithIV(ExValues.sampleIVcounter));

        IAeadCipher noPaddedDecryptor = decryptBuilder.BuildAeadCipher(AeadUsage.AAD_FIRST,
                    new MemoryInputStream(cipherTextData));
        noPaddedDecryptor.AadStream.Write(ExValues.additionalAuthenticatedDataA,0,
                                          ExValues.additionalAuthenticatedDataA.Length);
          
          
        byte[] plainTextData;
        using (Stream decrypterStream = noPaddedDecryptor.Stream)
        { plainTextData = Streams.ReadAll(decrypterStream); }    

        byte[][] plainMacData = { plainTextData, noPaddedDecryptor.GetMac().Collect() };
        
        return plainMacData;
        
    }

    /// <summary>
    /// Example 11 – GCM Mode Encryption no AAD
    /// </summary>    
    public static byte[] gcmEncrypt(FipsAes.Key key, byte[] plainTextData)
    {        
        IAeadCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IAeadCipherBuilder<IParameters<Algorithm>> encryptBuilder =
                    provider.CreateAeadEncryptorBuilder(FipsAes.Gcm.WithIV(ExValues.sampleIVcounter));

        MemoryOutputStream streamOutput = new MemoryOutputStream();
        
        ICipher noPaddedEncryptor = encryptBuilder.BuildCipher(streamOutput);
                                          
        using (Stream encrypterStream = noPaddedEncryptor.Stream)
        { encrypterStream.Write(plainTextData,0,plainTextData.Length); }    
          
          
        byte[] cipherTextData = streamOutput.ToArray();
        
        return cipherTextData;
    }
    
    /// <summary>
    /// Example 11 – GCM Mode Encryption no AAD
    /// </summary>    
    public static byte[] gcmDecrypt(FipsAes.Key key, byte[] cipherTextData)
    {        
        IAeadCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IAeadCipherBuilder<IParameters<Algorithm>> decryptBuilder =
                    provider.CreateAeadDecryptorBuilder(FipsAes.Gcm.WithIV(ExValues.sampleIVcounter));

        ICipher noPaddedDecryptor = decryptBuilder.BuildCipher(new MemoryInputStream(cipherTextData));
          
          
        byte[] plainTextData;
        using (Stream decrypterStream = noPaddedDecryptor.Stream)
        { plainTextData = Streams.ReadAll(decrypterStream); }    
       
        return plainTextData;
        
    }
   
    /// <summary>
    /// Example 12 – CCM Mode Encryption
    /// </summary>    
    public static byte[] ccmEncrypt(FipsAes.Key key, byte[] plainTextData)
    {
        IAeadCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IAeadCipherBuilder<IParameters<Algorithm>> encryptBuilder =
                    provider.CreateAeadEncryptorBuilder(FipsAes.Ccm.WithIV(ExValues.sampleIVcounter));

        MemoryOutputStream streamOutput = new MemoryOutputStream();
        
        ICipher noPaddedEncryptor = encryptBuilder.BuildCipher(streamOutput);
                                          
        using (Stream encrypterStream = noPaddedEncryptor.Stream)
        { encrypterStream.Write(plainTextData,0,plainTextData.Length); }    
          
          
        byte[] cipherTextData = streamOutput.ToArray();
        
        return cipherTextData;
    }  

    /// <summary>
    /// Example 12 – CCM Mode Decryption
    /// </summary>    
    public static byte[] ccmDecrypt(FipsAes.Key key, byte[] cipherTextData)
    {
        
        IAeadCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IAeadCipherBuilder<IParameters<Algorithm>> decryptBuilder =
                    provider.CreateAeadDecryptorBuilder(FipsAes.Ccm.WithIV(ExValues.sampleIVcounter));

        ICipher noPaddedDecryptor = decryptBuilder.BuildCipher(new MemoryInputStream(cipherTextData));
          
          
        byte[] plainTextData;
        using (Stream decrypterStream = noPaddedDecryptor.Stream)
        { plainTextData = Streams.ReadAll(decrypterStream); }    
       
        return plainTextData;
        
    }    
    
    /// <summary>
    /// Example 13 – CCM Mode Encryption with AAD
    /// </summary>    
    public static byte[] ccmEncryptAAD(FipsAes.Key key, byte[] plainTextData)
    {
        IAeadCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IAeadCipherBuilder<IParameters<Algorithm>> encryptBuilder =
                    provider.CreateAeadEncryptorBuilder(FipsAes.Ccm.WithIV(ExValues.sampleIVcounter));

        MemoryOutputStream streamOutput = new MemoryOutputStream();
        
        IAeadCipher noPaddedEncryptor = encryptBuilder.BuildAeadCipher(AeadUsage.AAD_FIRST, streamOutput);
        noPaddedEncryptor.AadStream.Write(ExValues.additionalAuthenticatedDataA,0,
                                          ExValues.additionalAuthenticatedDataA.Length);
                                          
        using (Stream encrypterStream = noPaddedEncryptor.Stream)
        { encrypterStream.Write(plainTextData,0,plainTextData.Length); }    
          
          
        byte[] cipherTextData = streamOutput.ToArray();
        
        return cipherTextData;
    }
    
    /// <summary>
    /// Example 13 – CCM Mode Deccryption with AAD
    /// </summary>    
    public static byte[] ccmDecryptAAD(FipsAes.Key key, byte[] cipherTextData)
    {
        IAeadCipherService provider = CryptoServicesRegistrar.CreateService(key); 
        IAeadCipherBuilder<IParameters<Algorithm>> decryptBuilder =
                    provider.CreateAeadDecryptorBuilder(FipsAes.Ccm.WithIV(ExValues.sampleIVcounter));

        IAeadCipher noPaddedDecryptor = decryptBuilder.BuildAeadCipher(AeadUsage.AAD_FIRST,
                    new MemoryInputStream(cipherTextData));
        noPaddedDecryptor.AadStream.Write(ExValues.additionalAuthenticatedDataA,0,
                                          ExValues.additionalAuthenticatedDataA.Length);
          
          
        byte[] plainTextData;
        using (Stream decrypterStream = noPaddedDecryptor.Stream)
        { plainTextData = Streams.ReadAll(decrypterStream); }    

        return plainTextData;
        
    }


    /// <summary>
    /// Example 33 – Wrapping without Padding. Note that the keyToWrap must be
    /// a multiple of 8 bytes (1/2 of 128 bits for AES)
    /// </summary>    
    public static byte[] wrapInputWithoutPadding(FipsAes.Key key, byte[] inputToWrap)
    {
        IAeadBlockCipherService provider = CryptoServicesRegistrar.CreateService(key);

        IKeyWrapper<FipsAes.WrapParameters> wrapper = provider.CreateKeyWrapper(FipsAes.KW);
        byte[] wrappedInput = wrapper.Wrap(inputToWrap).Collect();
        return wrappedInput;
    }

    public static byte[] unwrapInputWithoutPadding(FipsAes.Key key, byte[] cipherToUnwrap)
    {
        IAeadBlockCipherService provider = CryptoServicesRegistrar.CreateService(key);

        IKeyUnwrapper<FipsAes.WrapParameters> unwrapper = provider.CreateKeyUnwrapper(FipsAes.KW);
        byte[] unwrappedInput = unwrapper.Unwrap(cipherToUnwrap, 0, cipherToUnwrap.Length).Collect();
        return unwrappedInput;
    }    


    /// <summary>
    /// Example 34 – Wrapping without Padding. Note that the keyToWrap must be
    /// a multiple of 8 bytes (1/2 of 128 bits for AES)
    /// </summary>    
    public static byte[] wrapInputWithPadding(FipsAes.Key key, byte[] inputToWrap)
    {
        IAeadBlockCipherService provider = CryptoServicesRegistrar.CreateService(key);

        IKeyWrapper<FipsAes.WrapParameters> wrapper = provider.CreateKeyWrapper(FipsAes.KWP);
        byte[] wrappedInput = wrapper.Wrap(inputToWrap).Collect();
        return wrappedInput;
    }

    public static byte[] unwrapInputWithPadding(FipsAes.Key key, byte[] cipherToUnwrap)
    {
        IAeadBlockCipherService provider = CryptoServicesRegistrar.CreateService(key);

        IKeyUnwrapper<FipsAes.WrapParameters> unwrapper = provider.CreateKeyUnwrapper(FipsAes.KWP);
        byte[] unwrappedInput = unwrapper.Unwrap(cipherToUnwrap, 0, cipherToUnwrap.Length).Collect();
        return unwrappedInput;
    }

    /// <summary>
    /// Example 35 – wrapping and unwrapping a secure random key without padding
    /// </summary>    
    public static byte[] wrapKeyWithoutPadding(FipsAes.Key key, SymmetricSecretKey keyToWrap)
    {
        IAeadBlockCipherService provider = CryptoServicesRegistrar.CreateService(key);

        IKeyWrapper<FipsAes.WrapParameters> wrapper = provider.CreateKeyWrapper(FipsAes.KW);
        byte[] wrappedKey = wrapper.Wrap(keyToWrap.GetKeyBytes()).Collect();
        return wrappedKey;
    }
    
    public static SymmetricSecretKey unwrapKeyWithoutPadding(FipsAes.Key key, byte[] cipherToUnwrap)
    {
        IAeadBlockCipherService provider = CryptoServicesRegistrar.CreateService(key);

        IKeyUnwrapper<FipsAes.WrapParameters> unwrapper = provider.CreateKeyUnwrapper(FipsAes.KW);
        SymmetricSecretKey unwrappedKey = 
            new FipsAes.Key(unwrapper.Unwrap(cipherToUnwrap, 0, cipherToUnwrap.Length).Collect());
        return unwrappedKey;
    }
    

  
}
