using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Utilities.Test;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Utilities.IO;

using bcfipsin100.util;

public class DrbgExamples
{
    /// <summary>
    /// Example 1 – Creating a FIPS Approved SecureRandom
    /// </summary>
    public static FipsSecureRandom buildDrbg()
    {   
        BasicEntropySourceProvider entSource = new BasicEntropySourceProvider(new SecureRandom(), true);
        IDrbgBuilder<FipsSecureRandom> theDrbgBuilder = CryptoServicesRegistrar.CreateService(
                                                    FipsDrbg.Sha512HMac).FromEntropySource(entSource);                
        theDrbgBuilder.SetSecurityStrength(256);
        theDrbgBuilder.SetEntropyBitsRequired(256);
            
        return theDrbgBuilder.Build(ExValues.sampleNonce,false);
    }   

    /// <summary>
    /// Example 2 – Creating a FIPS Approved SecureRandom for Keys
    /// </summary>
    public static FipsSecureRandom buildDrbgForKeys()
    {
        BasicEntropySourceProvider entSource = new BasicEntropySourceProvider(new SecureRandom(), true);
        IDrbgBuilder<FipsSecureRandom> theDrbgBuilder = CryptoServicesRegistrar.CreateService(
                                                    FipsDrbg.Sha512HMac).FromEntropySource(entSource);                
        theDrbgBuilder.SetSecurityStrength(256);
        theDrbgBuilder.SetEntropyBitsRequired(256);
        theDrbgBuilder.SetPersonalizationString(ExValues.personalizationString);            

        return theDrbgBuilder.Build(ExValues.sampleNonce,true);
    }
    
    
    /// <summary>
    /// Example 3 – Configuring the Default SecureRandom
    /// </summary>
    public static void setDefaultDrbg()
    {
        BasicEntropySourceProvider entSource = new BasicEntropySourceProvider(new SecureRandom(), true);
        IDrbgBuilder<FipsSecureRandom> theDrbgBuilder = CryptoServicesRegistrar.CreateService(
                                                    FipsDrbg.Sha512HMac).FromEntropySource(entSource);                
        theDrbgBuilder.SetSecurityStrength(256);
        theDrbgBuilder.SetEntropyBitsRequired(256);
        CryptoServicesRegistrar.SetSecureRandom(theDrbgBuilder.Build(ExValues.sampleNonce, true));
    }
    
    public static void Main(String[] args)
    {
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("==========================");
        Console.WriteLine("Start Of DrbgExamples.Main");
        Console.WriteLine("==========================");
        

        byte[] drgb1Data = new byte[32];
        FipsSecureRandom drgb1 = buildDrbg();
        drgb1.NextBytes(drgb1Data);
        Console.WriteLine(string.Format("drgb1:[{0}]",string.Join(",",drgb1Data)));


        byte[] drgb2Data = new byte[32];
        FipsSecureRandom drgb2 = buildDrbgForKeys();
        drgb2.NextBytes(drgb2Data);
        Console.WriteLine(string.Format("drgb2:[{0}]",string.Join(",",drgb2Data)));

            
        setDefaultDrbg();
        /// Check we have set a default Drbg
        try
        {
            FipsSecureRandom currentDefaultSecureRandom = (FipsSecureRandom)CryptoServicesRegistrar.GetSecureRandom();
            Console.WriteLine("GREAT!!: FipsSecureRandom in CryptoServicesRegistrar is NOT null");
        }
        catch 
        {
            Console.WriteLine("error: current SecureRandom in CryptoServicesRegistrar is null");
        }
        
        Console.WriteLine("==========================");
        Console.WriteLine("End   Of DrbgExamples.Main");
        Console.WriteLine("==========================");
        Console.WriteLine();
        Console.WriteLine();

    }
}
