using System;
using System.Text;
using System.Reflection;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Utilities.Test;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Utilities.IO;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Crypto.Asymmetric;

using bcfipsin100.util;

public class DsaExamples
{
    
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("=================================");
        Console.WriteLine("Start Of DsaExamples.Main");
        Console.WriteLine("=================================");


        Console.Write("generating DSA key Pair without exisitng domain parameters ... ");
        AsymmetricKeyPair<AsymmetricDsaPublicKey, AsymmetricDsaPrivateKey> keyPair = 
            generateDSAKeyPair(2048,256);  
        Console.WriteLine("done");
        
        byte[] signature = generateDSASignature(keyPair.PrivateKey, ExValues.sampleSignatureMessage);        
        Console.WriteLine(string.Format("DSA SIGN:\tplainTextData=[{0}]\tHashTextData=[{1}]",
            System.Text.Encoding.Default.GetString(ExValues.sampleSignatureMessage),
            Hex.ToHexString(signature)));        
            
        bool result = verifyDSASignature(keyPair.PublicKey, signature, 
            Encoding.Default.GetBytes("Once Upon a midnight dreary, while I pondered weak and weary"));        
        Console.WriteLine(string.Format("DSA VERIFY:\tplainTextData=[{0}]\tVerification=[{1}]",
            System.Text.Encoding.Default.GetString(ExValues.sampleSignatureMessage),
           result ? "True" : "False"));
        
        
        Console.Write("generating DSA domain parameters ... ");
        DsaDomainParameters dsaDomainParams = generateDSADomainParameters(2048, 256);        
        Console.WriteLine("done");

        Console.Write("generating DSA key Pair with exisitng domain parameters ... ");
        keyPair = generateDSAKeyPairUsingParameters(dsaDomainParams);  
        Console.WriteLine("done");
        
        Console.WriteLine("=================================");
        Console.WriteLine("End   Of DsaExamples.Main");
        Console.WriteLine("=================================");
        Console.WriteLine();
        Console.WriteLine();        
    }
    
    public static FipsSecureRandom buildDrbg()
    {   
        BasicEntropySourceProvider entSource = new BasicEntropySourceProvider(new SecureRandom(), true);
        IDrbgBuilder<FipsSecureRandom> theDrbgBuilder = CryptoServicesRegistrar.CreateService(
                                                    FipsDrbg.Sha512HMac).FromEntropySource(entSource);                
        theDrbgBuilder.SetSecurityStrength(256);
        theDrbgBuilder.SetEntropyBitsRequired(25600);
            
        return theDrbgBuilder.Build(ExValues.sampleNonce,false);
    }   
    
    
    /// <summary>
    /// Example 22 – DSA Key Pair Generation
    /// </summary>    
    public static AsymmetricKeyPair<AsymmetricDsaPublicKey, AsymmetricDsaPrivateKey> 
                        generateDSAKeyPair(int L, int N)  
    {   
        FipsDsa.DomainGenParameters dsaDomainGenParams = new FipsDsa.DomainGenParameters(L,N);
        FipsDsa.DomainParametersGenerator dsaDomainParamsGen = 
            CryptoServicesRegistrar.CreateGenerator(dsaDomainGenParams,new SecureRandom());    
        DsaDomainParameters dsaDomainParams = dsaDomainParamsGen.GenerateDomainParameters();

        /*
        Console.WriteLine(string.Format("P=[{0}]\tQ=[{1}]\tG=[{2}]",
           dsaDomainParams.P.ToString(),
           dsaDomainParams.Q.ToString(),
           dsaDomainParams.G.ToString()));
        */

        FipsDsa.KeyGenerationParameters keyGenParameters =
            new FipsDsa.KeyGenerationParameters(dsaDomainParams);
                
        FipsDsa.KeyPairGenerator kpGen = 
            CryptoServicesRegistrar.CreateGenerator(keyGenParameters, new SecureRandom());

        return kpGen.GenerateKeyPair();
    }
    
    
    /// <summary>
    /// Example 23 – Signing with DSA
    /// </summary>    
    public static byte[] generateDSASignature(AsymmetricDsaPrivateKey key, byte[] message)
    {
        ISignatureFactoryService signatureFactoryProvider =
            CryptoServicesRegistrar.CreateService(key, new SecureRandom());

        ISignatureFactory<FipsDsa.SignatureParameters> signer = 
            signatureFactoryProvider.CreateSignatureFactory(FipsDsa.Dsa);

        IStreamCalculator<IBlockResult> calculator = signer.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().Collect();
    }

    /// <summary>
    /// Example 23 – Verifying with DSA 
    /// </summary>    
    public static bool verifyDSASignature(AsymmetricDsaPublicKey key, byte[] signature, byte[] message)
    {
        IVerifierFactoryService verifierFactoryProvider =
            CryptoServicesRegistrar.CreateService(key);

        IVerifierFactory<FipsDsa.SignatureParameters> verifier = 
            verifierFactoryProvider.CreateVerifierFactory(FipsDsa.Dsa);

        IStreamCalculator<IVerifier> calculator = verifier.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().IsVerified(signature);
    }

    /// <summary>
    /// Example 24 – Domain Parameter Generation
    /// </summary>    
    public static DsaDomainParameters generateDSADomainParameters(int L, int N)  
    {   
        FipsDsa.DomainGenParameters dsaDomainParams = new FipsDsa.DomainGenParameters(L,N);
        FipsDsa.DomainParametersGenerator dsaDomainParamGen = 
            CryptoServicesRegistrar.CreateGenerator(dsaDomainParams,new SecureRandom());    
        return  dsaDomainParamGen.GenerateDomainParameters();
    }    

    /// <summary>
    /// Example 25 – Generating Key Pairs using Parameters
    /// </summary>    
    public static AsymmetricKeyPair<AsymmetricDsaPublicKey, AsymmetricDsaPrivateKey> 
                        generateDSAKeyPairUsingParameters(DsaDomainParameters dsaDomainParams)  
    {   
        /*
        Console.WriteLine(string.Format("P=[{0}]\tQ=[{1}]\tG=[{2}]",
           dsaDomainParams.P.ToString(),
           dsaDomainParams.Q.ToString(),
           dsaDomainParams.G.ToString()));
        */

        FipsDsa.KeyGenerationParameters keyGenParameters =
            new FipsDsa.KeyGenerationParameters(dsaDomainParams);
                
        FipsDsa.KeyPairGenerator kpGen = 
            CryptoServicesRegistrar.CreateGenerator(keyGenParameters, new SecureRandom());

        return kpGen.GenerateKeyPair();
    }    

}
