using System;
using System.Text;
using System.Reflection;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Utilities.Test;
using Org.BouncyCastle.Crypto.Fips;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Utilities.IO;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Crypto.Asymmetric;

using bcfipsin100.util;

public class ECExamples
{
    
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("=================================");
        Console.WriteLine("Start Of ECExamples.Main");
        Console.WriteLine("=================================");

        AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey>  keyPair = generateECKeyPair();

        byte[] signature = generateECDSASignature(keyPair.PrivateKey,ExValues.sampleSignatureMessage);
        string  intSignature = new BigInteger(signature).ToString();       
        Console.WriteLine(string.Format("ECDSA signature=\t[{0}]",
            intSignature.Substring(1,20)+"..." + intSignature.Substring(intSignature.Length-20,20)));

        bool result = verifyECDSASignature(keyPair.PublicKey, signature,ExValues.sampleSignatureMessage); 
        Console.WriteLine(string.Format("ECDSA VERIFY\t\tplainTextData=[{0}]\tVerification=[{1}]",
           System.Text.Encoding.Default.GetString(ExValues.sampleSignatureMessage),
           result ? "True" : "False"));
           
        AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey> keyPairInitiator = 
            generateECKeyPairForDH(ExValues.DHFixedRandomA(1000));  
        ECDomainParameters domainParamsInitiator = keyPairInitiator.PublicKey.DomainParameters;
        Console.WriteLine(string.Format("\nECCDH Initiator Params\torder=[{0}]\tcofactor[{1}]",
            domainParamsInitiator.N.ToString(),domainParamsInitiator.H.ToString()));
            
        AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey> keyPairRecipient = 
            generateECKeyPairForDH(ExValues.DHFixedRandomA(1000));  
        ECDomainParameters domainParamsRecipient = keyPairRecipient.PublicKey.DomainParameters;
        Console.WriteLine(string.Format("ECCDH Recipient Params\torder=[{0}]\tcofactor[{1}]",
            domainParamsRecipient.N.ToString(),domainParamsRecipient.H.ToString()));
            
        byte[] initAgreement = initiatorAgreementBasic(keyPairInitiator.PrivateKey, keyPairRecipient.PublicKey);
        Console.WriteLine(string.Format("Initiator Basic Agreement:\tKey=[{0}]\tLen=[{1}]",
            string.Join("",ExValues.SubArray(initAgreement,0,20)) + "..." + 
            string.Join("",ExValues.SubArray(initAgreement,initAgreement.Length-20,20)),
            initAgreement.Length));
            
        byte[] recpAgreement = recipientAgreementBasic(keyPairRecipient.PrivateKey, keyPairInitiator.PublicKey);
        Console.WriteLine(string.Format("Recipient Basic Agreement:\tKey=[{0}]\tLen=[{1}]",
            string.Join("",ExValues.SubArray(recpAgreement,0,20)) + "..." + 
            string.Join("",ExValues.SubArray(recpAgreement,recpAgreement.Length-20,20)),
            recpAgreement.Length));

        initAgreement = initiatorAgreementWithX963KDF(keyPairInitiator.PrivateKey, keyPairRecipient.PublicKey,
                                                      ExValues.additionalAuthenticatedDataA);
        Console.WriteLine(string.Format("Initiator X963 Agreement:\tKey=[{0}]\tLen=[{1}]",
            string.Join("",initAgreement),initAgreement.Length));
            
        recpAgreement = recipientAgreementWithX963KDF(keyPairRecipient.PrivateKey, keyPairInitiator.PublicKey,
                                                     ExValues.additionalAuthenticatedDataA);
        Console.WriteLine(string.Format("Recipient X963 Agreement:\tKey=[{0}]\tLen=[{1}]",
            string.Join("",recpAgreement),recpAgreement.Length));

        Console.WriteLine("=================================");
        Console.WriteLine("End   Of ECExamples.Main");
        Console.WriteLine("=================================");
        Console.WriteLine();
        Console.WriteLine();        
    }
    
    /// <summary>
    /// Example 31 – EC Key Pair Generation using default Curves
    /// </summary>    
    public static AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey> 
        generateECKeyPair()  
    {   

        //Console.WriteLine(string.Format("P=[{0}]\tQ=[{1}]\tG=[{2}]",
        //   dsaDomainParams.P.ToString(),
        //   dsaDomainParams.Q.ToString(),
        //   dsaDomainParams.G.ToString()));


        FipsEC.KeyGenerationParameters keyGenParameters =
            new FipsEC.KeyGenerationParameters(FipsEC.DomainParams.P384);
                
        FipsEC.KeyPairGenerator kpGen = 
            CryptoServicesRegistrar.CreateGenerator(keyGenParameters, new SecureRandom());

        return kpGen.GenerateKeyPair();
    }      
    
    
    /// <summary>
    /// Example 32 – Signing with ECDSA
    /// </summary>    
    public static byte[] generateECDSASignature(AsymmetricECPrivateKey key, byte[] message)
    {
        ISignatureFactoryService signatureFactoryProvider =
            CryptoServicesRegistrar.CreateService(key, new SecureRandom());

        ISignatureFactory<FipsEC.SignatureParameters> signer = 
            signatureFactoryProvider.CreateSignatureFactory(FipsEC.Dsa);

        IStreamCalculator<IBlockResult> calculator = signer.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().Collect();
    }    
    
    
    /// <summary>
    /// Example 32 – Verifying with DSA 
    /// </summary>    
    public static bool verifyECDSASignature(AsymmetricECPublicKey key, byte[] signature, byte[] message)
    {
        IVerifierFactoryService verifierFactoryProvider =
            CryptoServicesRegistrar.CreateService(key);

        IVerifierFactory<FipsEC.SignatureParameters> verifier = 
            verifierFactoryProvider.CreateVerifierFactory(FipsEC.Dsa);

        IStreamCalculator<IVerifier> calculator = verifier.CreateCalculator();    

        Stream sOut = calculator.Stream;
        sOut.Write(message, 0, message.Length);
        sOut.Close();

        return calculator.GetResult().IsVerified(signature);
    }  

    /// <summary>
    /// Example 44 – EC Key Pair Generation using default Curves For DH
    /// </summary>    
    public static AsymmetricKeyPair<AsymmetricECPublicKey, AsymmetricECPrivateKey> 
        generateECKeyPairForDH(SecureRandom secRand)  
    {   

        //Console.WriteLine(string.Format("P=[{0}]\tQ=[{1}]\tG=[{2}]",
        //   dsaDomainParams.P.ToString(),
        //   dsaDomainParams.Q.ToString(),
        //   dsaDomainParams.G.ToString()));


        FipsEC.KeyGenerationParameters keyGenParameters =
            new FipsEC.KeyGenerationParameters(FipsEC.DomainParams.K409).For(FipsEC.Cdh);
                
        FipsEC.KeyPairGenerator kpGen = 
            CryptoServicesRegistrar.CreateGenerator(keyGenParameters, secRand);

        return kpGen.GenerateKeyPair();
    }   


    /// <summary>
    /// Example 45 – Basic ECCDH Key Agreement
    /// </summary>    
    public static byte[] initiatorAgreementBasic(
        AsymmetricECPrivateKey initiatorPrivate, AsymmetricECPublicKey recipientPublic)
    {
        IAgreementCalculatorService agreeFact =
            CryptoServicesRegistrar.CreateService(initiatorPrivate);

        IAgreementCalculator<FipsEC.AgreementParameters> agreement =
            agreeFact.CreateAgreementCalculator(FipsEC.Cdh);

        return agreement.Calculate(recipientPublic);    
    }
    
    public static byte[] recipientAgreementBasic(
        AsymmetricECPrivateKey recipientPrivate, AsymmetricECPublicKey initiatorPublic)
    {
        IAgreementCalculatorService agreeFact =
            CryptoServicesRegistrar.CreateService(recipientPrivate);

        IAgreementCalculator<FipsEC.AgreementParameters> agreement =
            agreeFact.CreateAgreementCalculator(FipsEC.Cdh);

        return agreement.Calculate(initiatorPublic);    
    }

    /// <summary>
    /// Example 46 – Basic ECCDH Key Agreement with a KDF
    /// </summary>
    public static byte[] initiatorAgreementWithX963KDF(AsymmetricECPrivateKey initiatorPrivate,
                            AsymmetricECPublicKey recipientPublic, byte[] userKeyingMaterial)
    {
        IAgreementCalculatorService agreeFact =
            CryptoServicesRegistrar.CreateService(initiatorPrivate);

        IAgreementCalculator<FipsEC.AgreementParameters> agreement =
            agreeFact.CreateAgreementCalculator(
                FipsEC.Cdh.WithKeyMaterialGenerator(
                    new FipsKdfKmg(FipsKdf.X963, FipsPrfAlgorithm.Sha384, userKeyingMaterial, 32)
                )
            );

        return agreement.Calculate(recipientPublic);    
    }
    
    public static byte[] recipientAgreementWithX963KDF(AsymmetricECPrivateKey recipientPrivate,   
                             AsymmetricECPublicKey initiatorPublic, byte[] userKeyingMaterial)
    {
        IAgreementCalculatorService agreeFact =
            CryptoServicesRegistrar.CreateService(recipientPrivate);

        IAgreementCalculator<FipsEC.AgreementParameters> agreement =
            agreeFact.CreateAgreementCalculator(
                FipsEC.Cdh.WithKeyMaterialGenerator(
                    new FipsKdfKmg(FipsKdf.X963, FipsPrfAlgorithm.Sha384, userKeyingMaterial, 32)
                )
            );

        return agreement.Calculate(initiatorPublic);    
    }
    
}

